import torch
import torch.nn as nn
import torch.optim as optim

import torch.nn.functional as F
import torchaudio
import tqdm

import argparse

import math

from torch.utils.data import random_split
from torch.utils.data import DataLoader

import AudioUtils
from data import AudioDataset
from generator import SISUGenerator
from discriminator import SISUDiscriminator

def perceptual_loss(y_true, y_pred):
    return torch.mean((y_true - y_pred) ** 2)

def discriminator_train(high_quality, low_quality, real_labels, fake_labels):
    optimizer_d.zero_grad()

    # Forward pass for real samples
    discriminator_decision_from_real = discriminator(high_quality[0])
    d_loss_real = criterion_d(discriminator_decision_from_real, real_labels)

    # Forward pass for fake samples (from generator output)
    generator_output = generator(low_quality[0])
    discriminator_decision_from_fake = discriminator(generator_output.detach())
    d_loss_fake = criterion_d(discriminator_decision_from_fake, fake_labels)

    # Combine real and fake losses
    d_loss = (d_loss_real + d_loss_fake) / 2.0

    # Backward pass and optimization
    d_loss.backward()
    nn.utils.clip_grad_norm_(discriminator.parameters(), max_norm=1.0)  # Gradient Clipping
    optimizer_d.step()

    return d_loss

def generator_train(low_quality, real_labels):
    optimizer_g.zero_grad()

    # Forward pass for fake samples (from generator output)
    generator_output = generator(low_quality[0])
    discriminator_decision = discriminator(generator_output)
    g_loss = criterion_g(discriminator_decision, real_labels)

    g_loss.backward()
    optimizer_g.step()
    return generator_output

# Init script argument parser
parser = argparse.ArgumentParser(description="Training script")
parser.add_argument("--generator", type=str, default=None,
                    help="Path to the generator model file")
parser.add_argument("--discriminator", type=str, default=None,
                    help="Path to the discriminator model file")

args = parser.parse_args()

# Check for CUDA availability
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print(f"Using device: {device}")

# Initialize dataset and dataloader
dataset_dir = './dataset/good'
dataset = AudioDataset(dataset_dir)

# ========= MULTIPLE =========

# dataset_size = len(dataset)
# train_size = int(dataset_size * .9)
# val_size = int(dataset_size-train_size)

#train_dataset, val_dataset = random_split(dataset, [train_size, val_size])

# train_data_loader = DataLoader(train_dataset, batch_size=1, shuffle=True)
# val_data_loader = DataLoader(val_dataset, batch_size=1, shuffle=True)

# ========= SINGLE =========

train_data_loader = DataLoader(dataset, batch_size=1, shuffle=True)

# Initialize models and move them to device
generator = SISUGenerator()
discriminator = SISUDiscriminator()

if args.generator is not None:
    generator.load_state_dict(torch.load(args.generator, weights_only=True))
if args.discriminator is not None:
    discriminator.load_state_dict(torch.load(args.discriminator, weights_only=True))

generator = generator.to(device)
discriminator = discriminator.to(device)

# Loss
criterion_g = nn.MSELoss()
criterion_d = nn.BCELoss()

# Optimizers
optimizer_g = optim.Adam(generator.parameters(), lr=0.0001, betas=(0.5, 0.999))
optimizer_d = optim.Adam(discriminator.parameters(), lr=0.0001, betas=(0.5, 0.999))

# Scheduler
scheduler_g = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_g, mode='min', factor=0.5, patience=5)
scheduler_d = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_d, mode='min', factor=0.5, patience=5)

def start_training():

    # Training loop

    # ========= DISCRIMINATOR PRE-TRAINING =========
    # discriminator_epochs = 1
    # for discriminator_epoch in range(discriminator_epochs):

    #     # ========= TRAINING =========
    #     for high_quality_clip, low_quality_clip in tqdm.tqdm(train_data_loader, desc=f"Epoch {discriminator_epoch+1}/{discriminator_epochs}"):
    #         high_quality_sample = high_quality_clip[0].to(device)
    #         low_quality_sample = low_quality_clip[0].to(device)

    #         scale = high_quality_clip[0].shape[2]/low_quality_clip[0].shape[2]

    #         # ========= LABELS =========
    #         batch_size = high_quality_sample.size(0)
    #         real_labels = torch.ones(batch_size, 1).to(device)
    #         fake_labels = torch.zeros(batch_size, 1).to(device)

    #         # ========= DISCRIMINATOR =========
    #         discriminator.train()
    #         discriminator_train(high_quality_sample, low_quality_sample, scale, real_labels, fake_labels)

    #     torch.save(discriminator.state_dict(), "models/discriminator-single-shot-pre-train.pt")

    generator_epochs = 5000
    for generator_epoch in range(generator_epochs):
        low_quality_audio = (torch.empty((1)), 1)
        high_quality_audio = (torch.empty((1)), 1)
        ai_enhanced_audio = (torch.empty((1)), 1)

        times_correct = 0

        # ========= TRAINING =========
        for high_quality_clip, low_quality_clip in tqdm.tqdm(train_data_loader, desc=f"Epoch {generator_epoch+1}/{generator_epochs}"):
        # for high_quality_clip, low_quality_clip in train_data_loader:
            high_quality_sample = (high_quality_clip[0].to(device), high_quality_clip[1])
            low_quality_sample = (low_quality_clip[0].to(device), low_quality_clip[1])

            # ========= LABELS =========
            batch_size = high_quality_clip[0].size(0)
            real_labels = torch.ones(batch_size, 1).to(device)
            fake_labels = torch.zeros(batch_size, 1).to(device)

            # ========= DISCRIMINATOR =========
            discriminator.train()
            discriminator_train(high_quality_sample, low_quality_sample, real_labels, fake_labels)

            # ========= GENERATOR =========
            generator.train()
            generator_output = generator_train(low_quality_sample, real_labels)

            # ========= SAVE LATEST AUDIO =========
            high_quality_audio = high_quality_clip
            low_quality_audio = low_quality_clip
            ai_enhanced_audio = (generator_output, high_quality_clip[1])

        #metric = snr(high_quality_audio[0].to(device), ai_enhanced_audio[0])
        #print(f"Generator metric {metric}!")
        #scheduler_g.step(metric)

        if generator_epoch % 10 == 0:
            print(f"Saved epoch {generator_epoch}!")
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-crap.wav", low_quality_audio[0][0].cpu(), high_quality_audio[1]) # <-- Because audio clip was resampled in data.py from original to crap and to original again.
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-ai.wav", ai_enhanced_audio[0][0].cpu(), ai_enhanced_audio[1])
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-orig.wav", high_quality_audio[0][0].cpu(), high_quality_audio[1])

        torch.save(discriminator.state_dict(), f"models/current-epoch-discriminator.pt")
        torch.save(generator.state_dict(), f"models/current-epoch-generator.pt")

    torch.save(discriminator.state_dict(), "models/epoch-5000-discriminator.pt")
    torch.save(generator.state_dict(), "models/epoch-5000-generator.pt")
    print("Training complete!")

start_training()
